<?php

declare(strict_types=1);

/*
 * Chill is a software for social workers
 *
 * For the full copyright and license information, please view
 * the LICENSE file that was distributed with this source code.
 */

namespace Chill\DocStoreBundle\Serializer\Normalizer;

use Chill\DocStoreBundle\Entity\StoredObject;
use Chill\DocStoreBundle\Security\Authorization\StoredObjectRoleEnum;
use Chill\DocStoreBundle\Security\Guard\JWTDavTokenProviderInterface;
use Symfony\Component\Routing\Generator\UrlGeneratorInterface;
use Symfony\Component\Serializer\Normalizer\NormalizerAwareInterface;
use Symfony\Component\Serializer\Normalizer\NormalizerAwareTrait;
use Symfony\Component\Serializer\Normalizer\NormalizerInterface;

/**
 * Class StoredObjectNormalizer.
 *
 * Normalizes a StoredObject entity to an array of data.
 */
final class StoredObjectNormalizer implements NormalizerInterface, NormalizerAwareInterface
{
    use NormalizerAwareTrait;
    public const ADD_DAV_SEE_LINK_CONTEXT = 'dav-see-link-context';
    public const ADD_DAV_EDIT_LINK_CONTEXT = 'dav-edit-link-context';

    public function __construct(
        private readonly JWTDavTokenProviderInterface $JWTDavTokenProvider,
        private readonly UrlGeneratorInterface $urlGenerator
    ) {}

    public function normalize($object, ?string $format = null, array $context = [])
    {
        /** @var StoredObject $object */
        $datas = [
            'datas' => $object->getDatas(),
            'filename' => $object->getFilename(),
            'id' => $object->getId(),
            'iv' => $object->getIv(),
            'keyInfos' => $object->getKeyInfos(),
            'title' => $object->getTitle(),
            'type' => $object->getType(),
            'uuid' => $object->getUuid(),
            'status' => $object->getStatus(),
            'createdAt' => $this->normalizer->normalize($object->getCreatedAt(), $format, $context),
            'createdBy' => $this->normalizer->normalize($object->getCreatedBy(), $format, $context),
        ];

        // deprecated property
        $datas['creationDate'] = $datas['createdAt'];

        $canDavSee = in_array(self::ADD_DAV_SEE_LINK_CONTEXT, $context['groups'] ?? [], true);
        $canDavEdit = in_array(self::ADD_DAV_EDIT_LINK_CONTEXT, $context['groups'] ?? [], true);

        if ($canDavSee || $canDavEdit) {
            $accessToken = $this->JWTDavTokenProvider->createToken(
                $object,
                $canDavEdit ? StoredObjectRoleEnum::EDIT : StoredObjectRoleEnum::SEE
            );

            $datas['_links'] = [
                'dav_link' => [
                    'href' => $this->urlGenerator->generate(
                        'chill_docstore_dav_document_get',
                        [
                            'uuid' => $object->getUuid(),
                            'access_token' => $accessToken,
                        ],
                        UrlGeneratorInterface::ABSOLUTE_URL,
                    ),
                    'expiration' => $this->JWTDavTokenProvider->getTokenExpiration($accessToken)->format('U'),
                ],
            ];
        }

        return $datas;
    }

    public function supportsNormalization($data, ?string $format = null)
    {
        return $data instanceof StoredObject && 'json' === $format;
    }
}
